import CryptoKit
import Security
import Foundation
import SwiftUI

/*
 digest = SHA512 (hmac & hkdf)
 hkdf salt = salt ^ othersalt
 cipher = aes gcm (salt = 12 bytes, tag = 16 bytes)
 message1: date(64bits) || salt(digest bit count) || kexpubkey (der) 
 message2: sig(m) || hmac(m) || m
 m = salt ^ othersalt || kexpubkey || otherkexpubkey
 */

typealias KexPrivKey = P384.KeyAgreement.PrivateKey
typealias KexPubKey = P384.KeyAgreement.PublicKey
typealias MyDigest = SHA512
typealias MyData = [UInt8]
let aesKeySize = 256/8
let info = "brenski".data(using:.ascii)!

enum CipherState {
    case initial
    case keycreated
    case secretbuilt
    case signaturesent
    //case finished
}

struct KeyCreated {
    let date:MyData
    let key:KexPrivKey
    let nonce:MyData
}

struct SecretBuilt {
    let mydate:MyData
    let otherdate:MyData
    let mynonce:MyData
    let othernonce:MyData
    let mypubkey:MyData
    let otherpubkey:MyData
    let key:SymmetricKey
    
    func message2payload() -> MyData {
        return mydate+otherdate+mynonce+othernonce+mypubkey+otherpubkey
    }
    func expected() -> MyData {
        return otherdate+mydate+othernonce+mynonce+otherpubkey+mypubkey
    }
}

struct SignatureSent {
    let expected: MyData
    let key:SymmetricKey
}

/*struct Finished {
    let key:SymmetricKey
}*/

struct Cipher {
    var state = CipherState.initial
    var keyCreated:KeyCreated? = nil
    var secretBuilt:SecretBuilt? = nil
    var signatureSent:SignatureSent? = nil
    //var finished:Finished? = nil
    
    mutating func reset() {
        keyCreated = nil
        secretBuilt = nil
        signatureSent = nil
        //finished = nil
        state = .initial
    }

    mutating func message1() throws -> String {
        guard case .initial = state else {
            reset()
            throw CryptoError("Invalid state")
        }
        let key = KexPrivKey(compactRepresentable: true)
        do {
            let date = timeBigEndian()
            let nonce = try createRandom(count:MyDigest.byteCount)
            let kexpub = key.publicKey.derRepresentation
            let total = date+nonce+kexpub
            state = .keycreated
            keyCreated = KeyCreated(date:date,key:key,nonce:nonce)
            return Data(total).base64EncodedString()
        }
        catch {
            reset()
            throw error
        }
    }

    mutating func handle(message1:String) throws -> Date {
        guard case .keycreated = state else {
            reset()
            throw CryptoError("Illegal state")
        }
        var input = message1
        trimString(&input)
        guard let data = Data(base64Encoded:input) else {
            reset()
            throw CryptoError("Invalid base64")
        }
        let nonceEnd = 8+MyDigest.byteCount
        //print(nonceEnd)
        if data.count<=nonceEnd {
            state = .initial
            throw CryptoError("Invalide message1 length")
        }
        let otherdate = MyData(data[0..<8])
        let date = try bigendianToDate(otherdate)
        let othernonce = MyData(data[8..<nonceEnd])
        let der = MyData(data[nonceEnd...])
        let mysha = MyDigest.hash(data:keyCreated!.nonce)
        let othersha = MyDigest.hash(data:othernonce)
        var myi = mysha.makeIterator()
        var otheri = othersha.makeIterator()
        var xor = MyData(repeating:0,count:MyDigest.byteCount)
        for n in 0..<MyDigest.byteCount {
            xor[n]=myi.next()!^otheri.next()!
        }
        do {
            let pub = try KexPubKey(derRepresentation: der)
            let secret = try keyCreated!.key.sharedSecretFromKeyAgreement(with: pub)
            let sym = secret.hkdfDerivedSymmetricKey(using: MyDigest.self, salt: xor, sharedInfo: info, outputByteCount: aesKeySize)
            //print(sym.withUnsafeBytes({buf in Data(buf).base64EncodedString()}))
            state = .secretbuilt
            secretBuilt = SecretBuilt(mydate:keyCreated!.date,otherdate:otherdate,
                                 mynonce:keyCreated!.nonce,othernonce:othernonce,mypubkey:MyData(keyCreated!.key.publicKey.derRepresentation),otherpubkey:der,key:sym)
            keyCreated = nil
            return date
        }
        catch {
            reset()
            throw error
        }
    }

    mutating func message2() throws -> String {
        guard case .secretbuilt = state else {
            reset()
            throw CryptoError("Invalid state")
        }
        let message2payload = secretBuilt!.message2payload()
        let sig = try sign(message2payload).rawRepresentation
        //print(sig.count)
        //print(sig.base64EncodedString())
        //let mac = HMAC<MyDigest>.authenticationCode(for: message2payload, using: sym)
        let cipher = try Cipher.cipher(sig,secretBuilt!.key)
        state = .signaturesent;
        signatureSent = SignatureSent(expected:secretBuilt!.expected(),key:secretBuilt!.key)
        //Swift.print(Data(secretBuilt!.message2payload()).base64EncodedString())
        secretBuilt = nil
        return cipher
    }

    mutating func handle(message2:String, directory:[DirEntry]) throws -> (String,SymmetricKey) {
        guard case .signaturesent = state else {
            reset()
            throw CryptoError("Invalid state")
        }
        do {
            let data = try Cipher.decipher(message2,signatureSent!.key)
            
        /* let s = data[0..<x]
        print(s.count)
        print(s.base64EncodedString()) */
            let sig = try Sig(rawRepresentation: data)
            //let mac = data[x...]
            guard let name = getSigner(data: signatureSent!.expected, sig: sig,directory: directory) else {
                throw CryptoError("Invalid signature")
            }
            state = .initial
            //finished = Finished(key:signatureSent!.key)
            return (name,key:signatureSent!.key)
        }
        catch {
            state = .initial
            throw error
        }
    }

    static func cipher(_ data:Data,_ sym:SymmetricKey) throws -> String {
        let nonce = AES.GCM.Nonce()
        let box = try AES.GCM.seal(data, using: sym, nonce: nonce)
        /*Swift.print(box.nonce.withUnsafeBytes { bytes in return
            Data(Array(bytes)).base64EncodedString()
        })*/
        //Swift.print(box.ciphertext.base64EncodedString())
        //Swift.print(box.tag.base64EncodedString())
        // nonce has 12 bytes, tag has 16 bytes
        // combined is nonce + ciphertext + tag
        return box.combined!.base64EncodedString()
    }
    
    

    static func decipher(_ message:String,_ sym:SymmetricKey) throws -> Data {
        var input = message
        trimString(&input)
        guard let data = Data(base64Encoded:input) else {
            throw CryptoError("Invalid base64")
        }
        //do {
        let box = try AES.GCM.SealedBox(combined:data)
        return try AES.GCM.open(box, using: sym)
    }
}
